/*
 * Copyright (c) 2002 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: driver_random.c,v 1.2 2002/05/23 12:09:15 zinnia Exp $
 */

#include "flip.h"

#define DRIVER_NAME "random"

#define LOWEST_BIT(x) ((x) & (~(x) + 1))
/* 
   ceiling 2^n

   see also: http://www.dd.iij4u.or.jp/~okuyamak/Information/flagcheck.html

*/
int ceiling(int m)
{
  int counter; int v, t;
  for(counter = 0, v = m; v != 0; counter++){
    t = LOWEST_BIT(v);
    v ^= t;
  }
  if(counter <= 1) return t;
  if(counter > 1) return t * 2;
}

/* 
  pseudo-uniform-random 

  x(n) = (A * x(n - 1) + C) Mod M

   M = 2^n;
   A = 5 mod 8
   C = 1 mod 2
*/
typedef struct {
  int now;
  int M;
  int M_ceiling; /* must be 2^M */
} pseudo_random_context;

#define PSEUDO_RANDOM_CONST_A 69
#define PSEUDO_RANDOM_CONST_C 21

void pseudo_random_init(pseudo_random_context* t, int M)
{
  t->now = 0;
  t->M = M;
  t->M_ceiling = ceiling(M);
  return;
}
int pseudo_random(pseudo_random_context* t)
{
  int n;
  n = t->now;
  do {
    n = (PSEUDO_RANDOM_CONST_A * n + PSEUDO_RANDOM_CONST_C) % t->M_ceiling;
  } while(n >= t->M);
  t->now = n;
  return n;
}


#define BLOCK_DEFAULT "8"

static int _timeout=0;

static int timeout()
{
  return _timeout;
}

static void flip(SDL_Surface* screen, SDL_Surface* bmp)
{
  Uint8* src; Uint8* dst;
  pseudo_random_context r;
  int count, x, y, offset, i;
  int block_x, block_y;

  timebase_context t;

  block_x = ABS(atoi(flip_get_property("x", BLOCK_DEFAULT)));
  block_y = ABS(atoi(flip_get_property("y", BLOCK_DEFAULT)));

  timebase_init(&t);
  pseudo_random_init(&r, bmp->w * bmp->h / (block_x * block_y));
#define GET_X(r) (((r).now % (bmp->w / block_x)) * block_x)
#define GET_Y(r) (((r).now / (bmp->w / block_x)) * block_y)

  src = bmp->pixels;
  dst = screen->pixels;

  for(count = 0; count < r.M; count++){
    pseudo_random(&r);
    x = GET_X(r);
    y = GET_Y(r);
    offset = y * screen->pitch + x * screen->format->BytesPerPixel;
    for(i = 0; i < block_y; i++){
      memcpy(dst + offset, src + offset, screen->format->BytesPerPixel * block_x);
      offset += screen->pitch;
    }

    if(timebase_elapse(&t, 3 * 1000 * count / r.M ) > 0) {
      //      SDL_UpdateRect(screen, x, y, block_x, block_y);
      SDL_UpdateRect(screen, 0, 0, 0, 0);
    } else {
      _timeout++;
    }
  }
  SDL_UpdateRect(screen, 0, 0, 0, 0);
 return;
}


static char* usage(FILE* fp)
{
  return DRIVER_NAME ":  random pixel swap\n"
    "  properties:\n"
    "    [x|y]=n: set block size to n (" BLOCK_DEFAULT ")";
}

driver_vector random_vector = {
  DRIVER_NAME,
  flip,
  timeout,
  usage,
};

